import readline
from appdirs import AppDirs
from pathlib import Path

from edpop_explorer.edpopxshell import EDPOPXShell


try:
    from colorama import just_fix_windows_console
    just_fix_windows_console()
except ImportError:
    pass

historyfile = Path(AppDirs('edpop-explorer', 'cdh').user_data_dir) / 'history'


def save_history() -> None:
    if not historyfile.parent.exists():
        historyfile.parent.mkdir(parents=True)
    readline.write_history_file(historyfile)


def main() -> None:
    if historyfile.exists():
        readline.read_history_file(historyfile)
    EDPOPXShell().cmdloop()
    save_history()


if __name__ == '__main__':
    main()



import cmd2
import math
import yaml
from typing import List, Optional, Type
from pygments import highlight
from pygments.lexers import TurtleLexer
from pygments.lexers.data import YamlLexer
from pygments.formatters import Terminal256Formatter

from edpop_explorer import Reader, Record, ReaderError
from edpop_explorer.readers import (
    FBTEEReader,
    GallicaReader,
    CERLThesaurusReader,
    HPBReader,
    VD16Reader,
    VD17Reader,
    VD18Reader,
    VDLiedReader,
    KBReader,
    STCNReader,
    SBTIReader,
    USTCReader,
    BnFReader,
)


class EDPOPXShell(cmd2.Cmd):
    intro = (
        'Welcome to the EDPOP explorer!\n'
        'Type <reader> <query> to perform a query.\n'
        'Type <reader> identifier <identifier> to retrieve a specific record.\n'
        'Type ‘help’ for all commands.\n'
    )
    prompt = '[edpop-explorer] # '
    reader: Optional[Reader] = None
    shown: int = 0
    RECORDS_PER_PAGE = 10

    def __init__(self):
        super().__init__()

        self.exact = False
        self.add_settable(cmd2.Settable(
            'exact', bool, 'use exact queries without preprocessing', self
        ))

    def get_record_from_argument(self, args) -> Optional[Record]:
        """Get the record requested by the user; show error message
        and return None if this fails"""
        if self.reader is None:
            self.perror('First perform an initial search')
            return
        try:
            # TODO: consider using argparse
            index = int(args) - 1
        except (TypeError, ValueError):
            self.perror('Please provide a valid number')
            return
        try:
            record = self.reader.records[index]
        except IndexError:
            self.perror('Please provide a record number that has been loaded')
            return
        return record

    def do_next(self, args) -> None:
        if self.reader is None:
            self.perror('First perform an initial search')
            return
        assert self.reader.number_of_results is not None
        assert self.reader.number_fetched is not None
        if self.shown >= self.reader.number_of_results:
            self.perror('All records have been shown')
        else:
            if self.reader.number_fetched - self.shown < self.RECORDS_PER_PAGE:
                self.reader.fetch()
            self.shown += self._show_records(self.reader.records,
                                             self.shown,
                                             self.RECORDS_PER_PAGE)

    def do_show(self, args) -> None:
        '''Show a normalized version of the record with the given number.'''
        record = self.get_record_from_argument(args)
        if record is not None:
            self.show_record(record)

    def show_record(self, record: Record) -> None:
        record.fetch()  # Necessary in case this is a lazy record
        self.poutput(cmd2.ansi.style_success(
            record, bold=True
        ))
        recordtype = str(record._rdf_class).rsplit('/',1)[1]
        self.poutput(f'Record type: {recordtype}')
        self.poutput
        if record.identifier:
            self.poutput(f'Identifier: {record.identifier}')
        if record.link:
            self.poutput('URL: ' + str(record.link))
        self.poutput(cmd2.ansi.style('Fields:', bold=True))
        for fieldname, _, _ in record._fields:
            fieldname_human = fieldname.capitalize().replace('_', ' ')
            # TODO: make a field iterator for Record
            value = getattr(record, fieldname)
            if value:
                if isinstance(value, list):
                    text = '\n' + '\n'.join([('  - ' + str(x)) for x in value])
                else:
                    text = str(value)
                self.poutput(
                    cmd2.ansi.style(f'- {fieldname_human}: ', bold=True) + text
                )

    def do_showrdf(self, args) -> None:
        '''Show an RDF representation of the record with the given number
        in Turtle format.'''
        record = self.get_record_from_argument(args)
        if record is None:
            return
        try:
            graph = record.to_graph()
            ttl = graph.serialize()
            highlighted = highlight(
                ttl, TurtleLexer(), Terminal256Formatter(style='vim')
            )
            self.poutput(highlighted)
        except ReaderError as err:
            self.perror('Cannot generate RDF: {}'.format(err))

    def do_showraw(self, args) -> None:
        '''Show the raw data of the record with the given number in the
        source catalog.'''
        record = self.get_record_from_argument(args)
        if record is None:
            return
        data = record.get_data_dict()
        yaml_data = yaml.dump(data, allow_unicode=True)
        highlighted = highlight(
            yaml_data, YamlLexer(), Terminal256Formatter(style='vim')
        )
        self.poutput(highlighted)

    def do_hpb(self, args) -> None:
        'CERL\'s Heritage of the Printed Book Database'
        self._query(HPBReader, args)

    def do_vd16(self, args) -> None:
        """Verzeichnis der im deutschen Sprachbereich erschienenen Drucke
        des 16. Jahrhunderts"""
        self._query(VD16Reader, args)

    def do_vd17(self, args) -> None:
        """Verzeichnis der im deutschen Sprachbereich erschienenen Drucke
        des 17. Jahrhunderts"""
        self._query(VD17Reader, args)

    def do_vd18(self, args) -> None:
        """Verzeichnis der im deutschen Sprachbereich erschienenen Drucke
        des 18. Jahrhunderts"""
        self._query(VD18Reader, args)

    def do_vdlied(self, args) -> None:
        """Verzeichnis der deutschsprachigen Liedflugschriften"""
        self._query(VDLiedReader, args)
    
    def do_bnf(self, args) -> None:
        """Bibliothèque nationale de France"""
        self._query(BnFReader, args)
    
    def do_gallica(self, args) -> None:
        'Gallica'
        self._query(GallicaReader, args)
    
    def do_ct(self, args) -> None:
        'CERL Thesaurus'
        self._query(CERLThesaurusReader, args)
    
    def do_stcn(self, args) -> None:
        'Short Title Catalogue Netherlands'
        self._query(STCNReader, args)
    
    def do_sbti(self, args) -> None:
        'Scottish Book Trade Index'
        self._query(SBTIReader, args)
    
    def do_fbtee(self, args) -> None:
        'French Book Trade in Enlightenment Europe'
        self._query(FBTEEReader, args)
    
    def do_ustc(self, args) -> None:
        'Universal Short Title Catalogue'
        self._query(USTCReader, args)
    
    def do_kb(self, args) -> None:
        'Koninklijke Bibliotheek'
        self._query(KBReader, args)

    def _show_records(self, records: List[Optional[Record]],
                      start: int,
                      limit=math.inf) -> int:
        """Show the records from start, with limit as the maximum number
        of records to show. Return the number of records shown."""
        total = len(records)
        remaining = total - start
        if remaining < 1:
            return 0
        # Determine count (the number of items to show)
        count = int(min(remaining, limit))
        digits = len(str(total))
        for i in range(start, start + count):
            print('{:{digits}} - {}'.format(
                i + 1, str(records[i]), digits=digits
            ))
        return count

    def _query(self, readerclass: Type[Reader], query: str):
        IDENTIFIER_PREFIX = "identifier "
        if query.startswith(IDENTIFIER_PREFIX):
            identifier = query[len(IDENTIFIER_PREFIX):]
            try:
                record = readerclass.get_by_id(identifier)
            except ReaderError as err:
                self.perror(err)
            else:
                self.show_record(record)
            return
        self.reader = readerclass()
        self.shown = 0
        try:
            if not self.exact:
                self.reader.prepare_query(query)
                self.pfeedback(
                    'Performing query: {}'.format(self.reader.prepared_query)
                )
            else:
                self.reader.set_query(query)
                self.pfeedback(
                    'Performing exact query: {}'.format(query)
                )
            self.reader.fetch()
        except ReaderError as err:
            self.perror('Error while fetching results: {}'.format(err))
            self.reader = None
            self.shown = 0
            return
        self.pfeedback(
            '{} records found.'.format(self.reader.number_of_results)
        )
        self.shown += self._show_records(
            self.reader.records, self.shown, self.RECORDS_PER_PAGE
        )



from typing import Dict, List

from rdflib import URIRef
from edpop_explorer import SRUReader, Record, BiographicalRecord, Field
from edpop_explorer.fields import LocationField


class CERLThesaurusReader(SRUReader):
    sru_url = 'https://data.cerl.org/thesaurus/_sru'
    sru_version = '1.2'
    CERL_LINK = 'https://data.cerl.org/thesaurus/{}'
    CTAS_PREFIX = 'http://sru.cerl.org/ctas/dtd/1.1:'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/cerlthesaurus'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/cerlthesaurus/"

    @classmethod
    def _get_acceptable_names(
            cls, namelist: List[Dict[str, str]]
    ) -> List[str]:
        names = []
        for name in namelist:
            if name['name'] in ['single', 'full']:
                names.append(name['text'])
        return names
    
    @classmethod
    def _convert_record(cls, sruthirecord: dict) -> Record:
        record = BiographicalRecord(from_reader=cls)
        record.identifier = sruthirecord['id']
        record.link = cls.CERL_LINK.format(record.identifier)
        record.data = sruthirecord

        # Add the names. Names are of different types: single, descriptive
        # (including the profession), full, inverted. inverted and
        # descriptive seem to be there always in addition to more
        # basic forms of the name. We will only choose the basic
        # forms for now. Names are both in headingForm (the default
        # display name) and variantForm (multiple variant names). We will
        # use these respectively for name and variantName.
        PREFIX = cls.CTAS_PREFIX
        headingform = sruthirecord.get(PREFIX + 'headingForm', None)
        if headingform and isinstance(headingform, list):
            names = cls._get_acceptable_names(headingform)
            if len(names):
                record.name = Field(names[0])
        # If no headingForm was defined, try display
        if not record.name:
            display = sruthirecord.get(PREFIX + 'display', None)
            if display:
                record.name = Field(display)
        variantform = sruthirecord.get(PREFIX + 'variantForm', None)
        if variantform and isinstance(variantform, list):
            names = cls._get_acceptable_names(variantform)
            record.variant_names = [Field(x) for x in names]

        # Add activityNote. This field can have only one value in CT.
        # NB: this data is very inconsistent and often includes other information
        # than somebody's activity - consider ignoring
        activitynote = sruthirecord.get(PREFIX + 'activityNote')
        if activitynote:
            record.activities = [Field(activitynote)]
        # Add biographicalData, which appears to be in all cases the years
        # that somebody was alive or that an entity existed
        biographicaldata = sruthirecord.get(PREFIX + 'biographicalData')
        if biographicaldata:
            record.timespan = Field(biographicaldata)
        # Add geographicalNote, which appears to be a country in all cases.
        # Add it to places of activity.
        geographicalnote = sruthirecord.get(PREFIX + 'geographicalNote')
        if geographicalnote:
            field = LocationField(geographicalnote)
            field.location_type = LocationField.COUNTRY
            record.places_of_activity = [field]

        return record

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query



from rdflib import Graph, Namespace, URIRef
from rdflib.term import Node
from typing import List, Optional, Tuple

from edpop_explorer import Field
from edpop_explorer.sparqlreader import (
    SparqlReader, BibliographicalRDFRecord
)


def _get_properties_from_iri(iri: str, properties: List[Node]) -> \
        Tuple[List[Node], Graph]:
    '''Get the first objects of the requested properties of a certain IRI
    as strings.'''
    subject_graph = Graph()
    subject_graph.parse(iri)
    objects: List[Node] = []
    for prop in properties:
        for obj in subject_graph.objects(URIRef(iri), prop):
            objects.append(obj)
    return objects, subject_graph


class STCNReader(SparqlReader):
    endpoint = 'http://data.bibliotheken.nl/sparql'
    filter = '?s schema:mainEntityOfPage/schema:isPartOf ' \
        '<http://data.bibliotheken.nl/id/dataset/stcn> .'
    name_predicate = '<http://schema.org/name>'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/stcn'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/stcn/"

    def __init__(self):
        super().__init__()

    @classmethod
    def convert_record(
        cls, graph: Graph, record: BibliographicalRDFRecord
    ) -> None:
        SCHEMA = Namespace('http://schema.org/')
        # First get the title and languages fields, which are simple
        # properties
        assert record.identifier is not None
        subject_node = URIRef(record.identifier)
        for name in graph.objects(subject_node, SCHEMA.name):
            record.title = Field(str(name))
            break
        record.languages = []
        for language in graph.objects(subject_node, SCHEMA.inLanguage):
            record.languages.append(Field(str(language)))
        # Now get the information from blank nodes
        record.contributors = []
        for author in graph.objects(subject_node, SCHEMA.author):
            name_field = None
            for name in graph.objects(author, SCHEMA.name):
                name_field = Field(str(name))
                # TODO: add role and authority record
            if name_field:
                record.contributors.append(name_field)
        for publication in graph.objects(subject_node, SCHEMA.publication):
            year_field = None
            for startDate in graph.objects(publication, SCHEMA.startDate):
                year_field = Field(str(startDate))
            if year_field:
                record.dating = year_field
            # TODO: publisher and location (not a blank node)
            published_by_iri = None
            for publishedBy in graph.objects(publication, SCHEMA.publishedBy):
                published_by_iri = str(publishedBy)
                break
            if published_by_iri:
                [name, location_node], pubgraph = _get_properties_from_iri(
                    published_by_iri, [SCHEMA.name, SCHEMA.location]
                )
                record.publisher_or_printer = Field(str(name))
                address_node = None
                for address in pubgraph.objects(location_node, SCHEMA.address):
                    address_node = address
                    break
                if address_node:
                    for addressLocality in pubgraph.objects(
                            address_node, SCHEMA.addressLocality
                    ):
                        record.place_of_publication = Field(
                            str(addressLocality)
                        )
                        break

    @classmethod
    def _create_lazy_record(
        cls, iri: str, name: Optional[str]=None
    ) -> BibliographicalRDFRecord:
        record = BibliographicalRDFRecord(cls)
        record.identifier = iri
        record.link = iri
        record.title = Field(name) if name else None
        return record



from typing import Optional

from rdflib import URIRef

from edpop_explorer import SRUMarc21BibliographicalReader, Marc21Data


class BnFReader(SRUMarc21BibliographicalReader):
    sru_url = 'http://catalogue.bnf.fr/api/SRU'
    sru_version = '1.2'
    HPB_LINK = 'http://hpb.cerl.org/record/{}'
    marcxchange_prefix = 'info:lc/xmlns/marcxchange-v2:'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/bnf'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/bnf/"
    _title_field_subfield = ('200', 'a')
    _alternative_title_field_subfield = ('500', 'a')
    _publisher_field_subfield = ('201', 'c')
    _place_field_subfield = ('210', 'a')
    _dating_field_subfield = ('210', 'd')
    _language_field_subfield = ('101', 'a')
    # TODO: add format etc

    @classmethod
    def transform_query(cls, query: str) -> str:
        return 'bib.anywhere all ({})'.format(query)

    @classmethod
    def _get_link(cls, data: Marc21Data) -> Optional[str]:
        # The link can be found in control field 003
        return data.controlfields.get('003', None)

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> str:
        return f'bib.anywhere all ("{identifier}")'

    @classmethod
    def _get_identifier(cls, data: Marc21Data) -> Optional[str]:
        return data.raw["id"]



def pytest_addoption(parser):
    parser.addoption('--requests', action='store_true', dest="requests",
                 default=False, help="enable tests with real API requests")

def pytest_configure(config):
    if not config.option.requests:
        setattr(config.option, 'markexpr', 'not requests')




from typing_extensions import override

import pytest

from rdflib import URIRef, RDF

from edpop_explorer import (
    Record,
    Reader,
    ReaderError,
    EDPOPREC,
    GetByIdBasedOnQueryMixin,
    NotFoundError,
)


class SimpleReader(Reader):
    """A simple reader which always yields 20 items which have as their
    ID the index number."""

    CATALOG_URIREF = URIRef("http://example.com/reader")
    IRI_PREFIX = "http://example.com/records/reader/"
    NUMBER_OF_ITEMS = 20

    @override
    def fetch_range(self, range_to_fetch: range) -> range:
        if range_to_fetch.stop > self.NUMBER_OF_ITEMS:
            range_to_fetch = range(range_to_fetch.start, self.NUMBER_OF_ITEMS)
        for i in range_to_fetch:
            self.records[i] = self.get_by_id(str(i))
        self.number_of_results = self.NUMBER_OF_ITEMS
        return range_to_fetch

    @classmethod
    @override
    def transform_query(cls, query):
        return query.capitalize()

    @classmethod
    @override
    def get_by_id(cls, identifier: str) -> Record:
        if int(identifier) not in range(cls.NUMBER_OF_ITEMS):
            raise NotFoundError
        record = Record(cls)
        record.identifier = identifier
        return record


class SimpleReaderNoIRIPrefix(SimpleReader):
    IRI_PREFIX = None


def test_catalog_to_graph():
    reader = SimpleReader()
    g = reader.catalog_to_graph()
    assert (reader.CATALOG_URIREF, RDF.type, EDPOPREC.Catalog) in g


def test_iri_to_identifier():
    iri = "http://example.com/records/reader/1"
    assert SimpleReader.iri_to_identifier(iri) == "1"


def test_iri_to_identifier_invalid():
    with pytest.raises(ReaderError):
        SimpleReader.iri_to_identifier("invalid/1")


def test_iri_to_identifier_and_vv_noprefixset():
    with pytest.raises(ReaderError):
        SimpleReaderNoIRIPrefix.iri_to_identifier(
            "http://example.com/records/reader/1"
        )
    with pytest.raises(ReaderError):
        SimpleReaderNoIRIPrefix.identifier_to_iri("1")


def test_identifier_to_iri():
    expected_iri = "http://example.com/records/reader/1"
    assert SimpleReader.identifier_to_iri("1") == expected_iri


def test_fetch():
    reader = SimpleReader()
    reader.set_query("test")  # Ignored by SimpleReader
    to_fetch = 8
    reader.fetch(to_fetch)
    assert reader.number_of_results == 20
    assert reader.number_fetched == to_fetch
    record = reader.get(5, False)
    assert isinstance(record, Record)
    # Fetch the next 8
    reader.fetch(to_fetch)
    record = reader.get(12, False)
    assert isinstance(record, Record)
    # Fetch the rest
    reader.fetch(to_fetch)
    assert reader.fetching_exhausted
    assert reader.number_fetched == reader.number_of_results
    with pytest.raises(NotFoundError):
        # Number 20 should not exist
        record = reader.get(20, False)


def test_get_allow_fetching():
    reader = SimpleReader()
    reader.set_query("test")
    record = reader.get(5, True)
    assert isinstance(record, Record)


def test_get_no_fetching():
    reader = SimpleReader()
    reader.set_query("test")
    with pytest.raises(NotFoundError):
        reader.get(5, False)


class SimpleReaderGetByIdBasedOnQuery(GetByIdBasedOnQueryMixin, SimpleReader):
    FETCH_ALL_AT_ONCE = True

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> str:
        return f"get {identifier}"

    @override
    def fetch_range(self, range_to_fetch: range) -> range:
        # To simplify, fetch all records at once
        if self.prepared_query == "get manymatching":
            self.records = {
                0: Record(self.__class__),
                1: Record(self.__class__),
            }
            self.number_of_results = 2
            return range(0, 2)
        elif self.prepared_query == "get nonematching":
            self.records = {}
            self.number_of_results = 0
            return range(0, 0)
        elif self.prepared_query.startswith("get "):
            record = Record(self.__class__)
            record.identifier = self.prepared_query[4:]
            self.records = {
                0: record,
            }
            self.number_of_results = 1
            return range(0, 1)


def test_getbyidbasedonquerymixin():
    record = SimpleReaderGetByIdBasedOnQuery.get_by_id("10")
    assert record.identifier == "10"


def test_getbyidbasedonquerymixin_multiplereturned():
    with pytest.raises(ReaderError):
        SimpleReaderGetByIdBasedOnQuery.get_by_id("manymatching")


def test_getbyidbasedonquerymixin_nonereturned():
    with pytest.raises(ReaderError):
        SimpleReaderGetByIdBasedOnQuery.get_by_id("nonematching")


def test_generate_identifier():
    reader = SimpleReader()
    reader.prepare_query("Hoi")
    identifier = reader.generate_identifier()
    # A new reader with the same query should have the same identifier. Now,
    # also perform a fetch.
    reader2 = SimpleReader()
    reader2.prepare_query("Hoi")
    reader2.fetch()
    identifier2 = reader2.generate_identifier()
    assert identifier == identifier2



"""RDF-related common functionality."""

from rdflib.namespace import Namespace
from rdflib import Graph, RDF, RDFS

EDPOPREC = Namespace('https://dhstatic.hum.uu.nl/edpop-records/latest/')
"""EDPOP Record Ontology"""

RELATORS = Namespace('http://id.loc.gov/vocabulary/relators/')
"""Library of Congress relators. See: https://id.loc.gov/vocabulary/relators.html"""


def bind_common_namespaces(graph: Graph) -> None:
    """Bind the RDF namespaces that are in use across this package to the
    specified graph.

    These are: RDF, RDFS, EDPOPREC."""

    graph.bind("rdf", RDF)
    graph.bind("rdfs", RDFS)
    graph.bind("edpoprec", EDPOPREC)




__all__ = [
    'EDPOPREC', 'RELATORS', 'bind_common_namespaces',
    'Field', 'FieldError', 'LocationField',
    'Reader', 'ReaderError', 'NotFoundError',
    'GetByIdBasedOnQueryMixin', 'BasePreparedQuery', 'PreparedQueryType',
    'Record', 'RawData', 'RecordError', 'BibliographicalRecord',
    'BiographicalRecord', 'LazyRecordMixin',
    'SRUReader',
    'Marc21Data', 'Marc21Field', 'Marc21BibliographicalRecord',
    'Marc21DataMixin', 'SRUMarc21Reader', 'SRUMarc21BibliographicalReader',
    'BIBLIOGRAPHICAL', 'BIOGRAPHICAL'
]

# Define here to avoid circular imports
# ruff: noqa
BIBLIOGRAPHICAL = "bibliographical"
BIOGRAPHICAL = "biographical"

from .rdf import EDPOPREC, RELATORS, bind_common_namespaces
from .fields import Field, FieldError, LocationField
from .reader import (
    Reader, ReaderError, GetByIdBasedOnQueryMixin, BasePreparedQuery,
    PreparedQueryType, NotFoundError
)
from .record import (
    Record, RawData, RecordError, BibliographicalRecord, BiographicalRecord,
    LazyRecordMixin
)
from .srureader import SRUReader
from .srumarc21reader import (
    Marc21Data, Marc21Field, Marc21BibliographicalRecord, Marc21DataMixin,
    SRUMarc21Reader, SRUMarc21BibliographicalReader
)




'''This package contains concrete subclasses of ``Reader``.'''

__all__ = [
    "BnFReader",
    "CERLThesaurusReader",
    "FBTEEReader",
    "GallicaReader",
    "HPBReader",
    "KBReader",
    "SBTIReader",
    "VD16Reader",
    "VD17Reader",
    "VD18Reader",
    "VDLiedReader",
    "STCNReader",
    "USTCReader",
    "ALL_READERS",
]

from edpop_explorer import Reader

from .bnf import BnFReader
from .cerl_thesaurus import CERLThesaurusReader
from .fbtee import FBTEEReader
from .gallica import GallicaReader
from .hpb import HPBReader
from .kb import KBReader
from .sbtireader import SBTIReader
from .stcn import STCNReader
from .ustc import USTCReader
from .vd import VD16Reader, VD17Reader, VD18Reader, VDLiedReader

import sys
from typing import List, Type


def _get_all_readers() -> List[Type[Reader]]:
    """Create a list of all reader classes included in this package."""
    all_names = __all__.copy()
    all_names.pop()  # Remove "ALL_READERS" itself
    all_readers: List[Type[Reader]] = []
    for name in all_names:
        cls = getattr(sys.modules[__name__], name, None)
        if cls is not None:
            all_readers.append(cls)
    return all_readers

ALL_READERS = _get_all_readers()




from abc import ABC, abstractmethod
from typing import Type, Tuple, Union, Optional, List, TYPE_CHECKING
from rdflib.term import Node
from rdflib import URIRef, Graph, BNode, RDF, Literal

from edpop_explorer import (
    EDPOPREC, Field, BIBLIOGRAPHICAL, BIOGRAPHICAL, bind_common_namespaces
)

if TYPE_CHECKING:
    from edpop_explorer import Reader


class RawData(ABC):
    """Base class to store raw original data of a record. Only defines
    an abstract method ``to_dict``.
    """

    @abstractmethod
    def to_dict(self) -> dict:
        """Give a ``dict`` representation of the raw data."""
        pass


class RecordError(Exception):
    pass


class Record:
    """Python representation of edpoprec:Record.

    This base class provides some basic attributes, an infrastructure
    to define fields and a method to convert the record to RDF.
    While this is a non-abstract base class, no fields are
    defined here -- these should be added in subclasses. ``Record``
    and its subclasses should be created by calling the constructor
    with the ``Reader`` class as the ``from_reader`` parameter
    and by setting the ``data``, ``link``, ``identifier`` and
    ``subject_node`` attributes (all are optional but recommended),
    as well as the fields that are defined by the subclass.
    fields are set using the attribute with the same name:
    set them to an instance of ``Field`` or to ``None``. The
    basic attributes and the fields are ``None`` by default.

    Subclasses should override the ``_rdf_class`` attribute to
    the corresponding RDF class. They should define additional 
    fields by adding additional public attributes defaulting
    to ``None`` and by registring them in the ``_fields`` attribute.
    For registring, a constructor ``__init__`` should be defined
    that first calls the parent's constructor and then adds the
    fields by adding tuples to ``_fields`` in the form
    ``('<attribute-name>', EDPOPREC.<rdf-property-name>,
    <Field class name>)``.
    """
    #: The raw original data of a record.
    data: Union[None, dict, RawData] = None
    _fields: List[Tuple[str, URIRef, Type[Field]]]
    _rdf_class: Node = EDPOPREC.Record
    link: Optional[str] = None
    '''A user-friendly link where the user can find the record.'''
    identifier: Optional[str] = None
    '''Unique identifier used by the source catalog.'''
    from_reader: Type["Reader"]
    '''The subject node, which will be used to convert the record to 
    RDF. This is a blank node by default.'''
    _graph: Optional[Graph] = None
    _bnode: Optional[BNode] = None

    def __init__(self, from_reader: Type["Reader"]):
        self._fields = []
        self.from_reader = from_reader

    def to_graph(self) -> Graph:
        '''Return an RDF graph for this record.'''
        self.fetch()
        g = Graph()
        
        # Set basic properties
        rdfclass = EDPOPREC.Record
        if self.from_reader:
            if self.from_reader.READERTYPE == BIOGRAPHICAL:
                rdfclass = EDPOPREC.BiographicalRecord
            elif self.from_reader.READERTYPE == BIBLIOGRAPHICAL:
                rdfclass = EDPOPREC.BibliographicalRecord
        g.add((
            self.subject_node,
            RDF.type,
            rdfclass
        ))
        if self.from_reader is not None and \
                self.from_reader.CATALOG_URIREF is not None:
            g.add((
                self.subject_node,
                EDPOPREC.fromCatalog,
                self.from_reader.CATALOG_URIREF
            ))
        if self.identifier:
            g.add((
                self.subject_node,
                EDPOPREC.identifier,
                Literal(self.identifier)
            ))
        if self.link:
            g.add((
                self.subject_node,
                EDPOPREC.publicURL,
                Literal(self.link)
            ))

        # Put all fields from self.FIELDS in the graph by accessing
        # the associated attributes or properties. If they contain a
        # list of Fields, repeat them in RDF.
        assert isinstance(self._fields, list)
        for field in self._fields:
            attrname, propref, fieldclass = field
            if not issubclass(fieldclass, Field):
                raise RecordError(
                    f"{attrname} in {self.__class__}.FIELDS is defined as being "
                    "of type {fieldclass}, but this type does not inherit from "
                    "Field."
                )
            value_or_values = getattr(self, attrname, None)
            if isinstance(value_or_values, list):
                values = value_or_values
            else:
                values = [value_or_values]
            for value in values:
                if value is None:
                    # self does not have the attribute or the attribute's value
                    # is None; ignore
                    continue
                if not isinstance(value, fieldclass):
                    raise RecordError(
                        f"{attrname} attribute is of type {type(value)} while an "
                        "instance of {fieldclass} was expected."
                    )
                partial_graph = value.to_graph()
                g.add((self.subject_node, propref, value.subject_node))
                g += partial_graph

        # Set namespace prefixes
        bind_common_namespaces(g)

        return g

    def get_data_dict(self) -> Optional[dict]:
        """Convenience function to get the record's raw data as a ``dict``,
        or ``None`` if it is not available."""
        self.fetch()
        if isinstance(self.data, RawData):
            return self.data.to_dict()
        else:
            # self.data should be dict or None
            return self.data

    def __str__(self):
        if self.identifier:
            return f'{self.__class__} object ({self.identifier})'
        else:
            return f'{self.__class__} object'

    def fetch(self) -> None:
        '''Fetch the full contents of the record if this record works with
        lazy loading (i.e., if the record's class derives from
        ``RDFRecordMixin``). If the record is not lazy, this method does
        nothing.'''
        pass

    @property
    def iri(self) -> Optional[str]:
        '''A stable IRI based on the `identifier` attribute. `None` if
        the `identifier` attribute is not set.'''
        if self.identifier:
            return self.from_reader.identifier_to_iri(self.identifier)
        else:
            return None

    @property
    def subject_node(self) -> Node:
        '''A subject node based on the `identifier` attribute. If the 
        `identifier` attribute is not set, a blank node.'''
        iri = self.iri
        if iri is not None:
            return URIRef(iri)
        else:
            # IRI is not available; return a consistent blank node
            if not self._bnode:
                self._bnode = BNode()
            return self._bnode


class BibliographicalRecord(Record):
    '''Python representation of edpoprec:BibliographicalRecord.

    This subclass adds fields that are specific for bibliographical
    records.
    '''
    _rdf_class = EDPOPREC.BibliographicalRecord
    title: Optional[Field] = None
    alternative_title: Optional[Field] = None
    contributors: Optional[List[Field]] = None
    publisher_or_printer: Optional[Field] = None
    place_of_publication: Optional[Field] = None
    dating: Optional[Field] = None
    languages: Optional[List[Field]] = None
    extent: Optional[Field] = None
    size: Optional[Field] = None
    physical_description: Optional[Field] = None

    def __init__(self, from_reader: Type["Reader"]):
        super().__init__(from_reader)
        assert isinstance(self._fields, list)
        self._fields += [
            ('title', EDPOPREC.title, Field),
            ('alternative_title', EDPOPREC.alternativeTitle, Field),
            ('contributors', EDPOPREC.contributor, Field),
            ('publisher_or_printer', EDPOPREC.publisherOrPrinter, Field),
            ('place_of_publication', EDPOPREC.placeOfPublication, Field),
            ('dating', EDPOPREC.dating, Field),
            ('languages', EDPOPREC.language, Field),
            ('extent', EDPOPREC.extent, Field),
            ('size', EDPOPREC.size, Field),
            ('physical_description', EDPOPREC.physicalDescription, Field),
        ]

    def __str__(self) -> str:
        if self.title:
            return str(self.title)
        else:
            return super().__str__()


class BiographicalRecord(Record):
    '''Python representation of edpoprec:BiographicalRecord.

    This subclass adds fields that are specific for biographical records.
    '''
    _rdf_class = EDPOPREC.BiographicalRecord
    name: Optional[Field] = None
    variant_names: Optional[List[Field]] = None
    place_of_birth: Optional[Field] = None
    place_of_death: Optional[Field] = None
    places_of_activity: Optional[List[Field]] = None
    timespan: Optional[Field] = None
    activities: Optional[List[Field]] = None

    def __init__(self, from_reader: Type["Reader"]):
        super().__init__(from_reader)
        assert isinstance(self._fields, list)
        self._fields += [
            ('name', EDPOPREC.title, Field),
            ('variant_names', EDPOPREC.variantName, Field),
            ('place_of_birth', EDPOPREC.placeOfBirth, Field),
            ('place_of_death', EDPOPREC.placeOfDeath, Field),
            ('places_of_activity', EDPOPREC.placeOfActivity, Field),
            ('timespan', EDPOPREC.timespan, Field),
            ('activities', EDPOPREC.activity, Field),
        ]

    def __str__(self) -> str:
        if self.name:
            return str(self.name)
        else:
            return super().__str__()


class LazyRecordMixin(ABC):
    '''Abstract mixin that adds an interface for lazy loading to a Record.

    To use, implement the ``fetch()`` method and make sure that it fills
    the record's ``data`` attributes and its Fields and that the 
    ``fetched`` attribute is set to ``True``.'''
    fetched: bool = False
    
    @abstractmethod
    def fetch(self) -> None:
        pass



from rdflib import URIRef
from typing import Optional

from edpop_explorer import (
    SRUMarc21BibliographicalReader, Marc21Data, BIBLIOGRAPHICAL
)


class HPBReader(SRUMarc21BibliographicalReader):
    sru_url = 'http://sru.k10plus.de/hpb'
    sru_version = '1.1'
    HPB_LINK = 'http://hpb.cerl.org/record/{}'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/hpb'
    )
    READERTYPE = BIBLIOGRAPHICAL
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/hpb/"

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> str:
        return f"pica.cid={identifier}"

    @classmethod
    def _get_identifier(cls, data:Marc21Data) -> Optional[str]:
        # The record id can be found in field 035 in subfield a starting
        # with (CERL), like this: (CERL)HU-SzSEK.01.bibJAT603188.
        # The URI can then be created using HPB_URI.
        # HPB records have field 035 two times.
        fields035 = data.get_fields('035')
        for field in fields035:
            if 'a' in field.subfields and \
                    field.subfields['a'].startswith('(CERL)'):
                return field.subfields['a'][len('(CERL)'):]


    @classmethod
    def _get_link(cls, data: Marc21Data) -> Optional[str]:
        identifier = cls._get_identifier(data)
        if identifier is not None:
            return cls.HPB_LINK.format(identifier)
        else:
            return None



from dataclasses import dataclass
from typing import List, Union

from edpop_explorer import BasePreparedQuery


@dataclass
class SQLPreparedQuery(BasePreparedQuery):
    where_statement: str
    arguments: List[Union[str, int]]



from rdflib import URIRef
from edpop_explorer import SRUReader, BibliographicalRecord, Field
from typing import Optional
import re
import requests
import xmltodict


def _force_list(data) -> list:
    if isinstance(data, list):
        return data
    elif data is None:
        return []
    else:
        return [data]

def _force_string(data) -> Optional[str]:
    '''Transform data into one string or None. Can be used if a single 
    string is expected, but if there is a possibility that it is a
    list.'''
    if isinstance(data, list):
        return ' ; '.join([str(x) for x in data])
    elif data is None:
        return None
    else:
        return str(data)


class GallicaReader(SRUReader):
    sru_url = 'https://gallica.bnf.fr/SRU'
    sru_version = '1.2'
    CERL_LINK = 'https://data.cerl.org/thesaurus/{}'
    CTAS_PREFIX = 'http://sru.cerl.org/ctas/dtd/1.1:'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/gallica'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/gallica/"
    DOCUMENT_API_URL = "https://gallica.bnf.fr/services/OAIRecord?ark={}"
    IDENTIFIER_PREFIX = "https://gallica.bnf.fr/"

    @classmethod
    def _convert_record(cls, sruthirecord: dict) -> BibliographicalRecord:
        record = BibliographicalRecord(cls)
        # identifier field contains visitable Gallica URL and possibly
        # also other types of identifier. In the first case we get it as a
        # string from sruthi, in the latter case as a list of strings.
        # Take the first string starting with https:// as the identifier
        # and as the link.
        identifiers = _force_list(sruthirecord.get('identifier', None))
        for identifier in identifiers:
            if identifier.startswith(cls.IDENTIFIER_PREFIX):
                record.identifier = identifier[len(cls.IDENTIFIER_PREFIX):]
                record.link = identifier
        record.data = {}
        for key in sruthirecord:
            if key in ['schema', 'id']:
                continue
            showkey: str = key.replace(cls.CTAS_PREFIX, 'ctas:')
            record.data[showkey] = sruthirecord[key]
        record.data = sruthirecord
        title = _force_string(sruthirecord.get('title', None))
        if title:
            record.title = Field(title)
        creators = _force_list(sruthirecord.get('creator', None))
        record.contributors = [Field(x) for x in creators]
        dating = _force_string(sruthirecord.get('date', None))
        if dating:
            record.dating = Field(dating)
        languages = _force_list(sruthirecord.get('language', None))
        record.languages = [Field(x) for x in languages]
        publisher = _force_string(sruthirecord.get('publisher', None))
        if publisher:
            record.publisher_or_printer = Field(publisher)

        # Format is more complicated: this is a list and generally contains
        # the number of views, the MIME type and the extent.
        # Try finding the extent by filtering out the other two.
        # This seems to work correctly.
        format_strings = _force_list(sruthirecord.get('format', None))
        for formatstr in format_strings:
            if not (formatstr.startswith('Nombre total de vues') or
                    re.match('$[a-z]+/[a-z]+^', formatstr)):
                record.extent = Field(formatstr)
                break

        return record

    @classmethod
    def get_by_id(cls, identifier: str) -> BibliographicalRecord:
        # Getting by id works via another interface (a simple XML API), but the 
        # returned data is the same in a slightly different format. Hence,
        # convert it to JSON just like sruthi does and extract the right piece
        # of data.
        url = cls.DOCUMENT_API_URL.format(identifier)
        res = requests.get(url, headers={"accept": "application/xml"})
        response_as_dict = xmltodict.parse(
            res.text,
            dict_constructor=dict,
            process_namespaces=True,
            namespaces={  # Remove these namespace prefixes
                "http://purl.org/dc/elements/1.1/": None,
                "http://www.openarchives.org/OAI/2.0/oai_dc/": None,
            },
            attr_prefix="",
            cdata_key="text",
        )
        data = response_as_dict["results"]["notice"]["record"]["metadata"]["dc"]
        # The returned XML has elements with attributes, while these attributes
        # are missing from the XML that is sent back by the SRU interface.
        # An attribute-less element is represented as a simple string by 
        # xmltodict, while an attribute with elements is represented as a 
        # dict where the contents is in the value of "text". Replace these
        # dicts with simple strings. (Not a very clean solution but refactoring
        # is not worth the time at this point.)
        for key in data:
            value = data[key]
            if isinstance(value, list):
                for index in range(len(value)):
                    item = value[index]
                    if isinstance(item, dict):
                        value[index] = item["text"]
        return cls._convert_record(data)

    @classmethod
    def transform_query(cls, query: str) -> str:
        return 'gallica all {}'.format(query)



import pytest
from rdflib import Literal, URIRef
from typing import Optional, List

from edpop_explorer import Record, Reader, Field, RecordError
from edpop_explorer import EDPOPREC
from edpop_explorer.record import BiographicalRecord


class SimpleReader(Reader):
    CATALOG_URIREF = URIRef('http://example.com/reader')
    IRI_PREFIX = "http://example.com/records/reader/"


class SimpleRecord(Record):
    _rdf_class = EDPOPREC.SimpleRecord
    testfield: Optional[Field] = None
    multiplefield: Optional[List[Field]] = None

    def __init__(self, from_reader):
        super().__init__(from_reader)
        self._fields.extend([
            ('testfield', EDPOPREC.testField, Field),
            ('multiplefield', EDPOPREC.multipleField, Field)
        ])


@pytest.fixture
def basic_record():
    record = SimpleRecord(SimpleReader)
    record.link = 'http://example.com'
    record.identifier = '123'
    return record


def test_iri(basic_record: SimpleRecord):
    assert basic_record.iri == "http://example.com/records/reader/123"


def test_iri_empty(basic_record: SimpleRecord):
    basic_record.identifier = None
    assert basic_record.iri is None


def test_subject_node(basic_record: SimpleRecord):
    assert basic_record.subject_node == \
            URIRef("http://example.com/records/reader/123")


def test_to_graph_empty():
    # Test if it works with an empty record
    record = Record(SimpleReader)
    g = record.to_graph()
    assert (
        record.subject_node, EDPOPREC.fromCatalog, SimpleReader.CATALOG_URIREF
    ) in g
    

def test_to_graph_basic_attributes(basic_record):
    g = basic_record.to_graph()
    assert (
        basic_record.subject_node, EDPOPREC.publicURL, Literal(basic_record.link)
    ) in g
    assert (
        basic_record.subject_node, EDPOPREC.identifier, Literal(basic_record.identifier)
    ) in g


def test_to_graph_empty_field(basic_record):
    # If testfield is None (the default), it should be absent from the graph
    g = basic_record.to_graph()
    assert (basic_record.subject_node, EDPOPREC.testField, None) not in g


def test_to_graph_field_normal_value(basic_record):
    basic_record.testfield = Field('test')
    g = basic_record.to_graph()
    assert (basic_record.subject_node, EDPOPREC.testField, None) in g
    

def test_to_graph_string_in_field(basic_record):
    basic_record.testfield = 'test'  # type: ignore
    with pytest.raises(RecordError):
        basic_record.to_graph()
    
def test_to_graph_field_multiple_values(basic_record):
    # Try a field that accepts multiple values
    basic_record.multiplefield = [
        Field('v1'), Field('v2')
    ]
    g = basic_record.to_graph()
    assert len(list(
        g.objects(basic_record.subject_node, EDPOPREC.multipleField)
    )) == 2

def test_biographicalrecord():
    # Basic test to check if class definition is sane; the logic should be
    # handled by super class
    record = BiographicalRecord(SimpleReader)
    record.to_graph()

def test_biographicalrecord_str():
    record = BiographicalRecord(SimpleReader)
    personname = "Person"
    record.name = Field(personname)
    assert str(record) == personname



"""This module defines ``Field`` and a number of subclasses for specialized
fields. 
"""

from typing import Optional, Callable, List, Tuple
from rdflib import Graph, Literal, BNode, RDF, URIRef
from rdflib.term import Node

from edpop_explorer import EDPOPREC

DATATYPES = {
    'string': {
        'input_type': str,
        'converter': (lambda x: Literal(x)),
    },
    'boolean': {
        'input_type': bool,
        'converter': (lambda x: Literal(x)),
    },
    'edtf': {
        'input_type': str,
        'converter': (
            lambda x: Literal(
                x, datatype=URIRef("http://id.loc.gov/datatypes/edtf")
            )
        )
    },
    'uriref': {
        'input_type': URIRef,
        'converter': lambda x: x,
    },
}


class FieldError(Exception):
    pass


class Field:
    """Python representation of edpoprec:Field.

    This base class has two user-defined subfields: ``original_text`` (which is
    required and should be passed to the constructor), and ``unknown``.
    User-defined subfields are simple object attributes and can be accessed
    directly. In addition, this base class defines an automatic
    subfield ``normalized_text``, which is a read-only property that is only
    available if normalization is supported by the field -- this is
    not the case for this base class. In those cases, it is still possible to
    set this field using the ``set_normalized_text`` method.
    Except ``original_text``, all subfields are optional and are None by default.
    Use ``to_graph()`` to obtain an RDF graph. The subject node is by default 
    a blank node, but this may be overridden by setting the subject_node 
    attribute.

    Subclasses should override the ``_rdf_class`` attribute to the corresponding
    RDF class. Subclasses can define additional subfields by adding additional
    public attributes and by registring them in the ``SUBFIELDS`` constant
    attribute. For registring, a constructor ``__init__`` should be defined that
    first calls the parent's constructor and then adds the subfields one
    by one using ``self.SUBFIELDS.append(('<attribute-name>',
    EDPOPREC.<rdf-property-name>, '<datatype>'))``, where <datatype> is any
    of the datatypes defined in the ``DATATYPES`` constant of this module.
    Subclasses may furthermore define the ``_normalized_text`` private 
    method."""
    #: Subfield -- text of this field according to the original record.
    original_text: str
    #: This field's subject node if converted to RDF. This is a blank node
    #: by default.
    subject_node: Node
    _subfields: List[Tuple[str, URIRef, str]]
    _normalized_text: Optional[str] = None
    #: Subfield -- indicates whether the value of this field is explicitly
    #: marked as unknown in the original record.
    unknown: Optional[bool] = None
    #: Subfield -- may contain the URI of an authority record
    authority_record: Optional[str] = None
    _create_normalized_text: Optional[Callable] = None
    _rdf_class: Node = EDPOPREC.Field
    
    def __init__(self, original_text: str) -> None:
        if not isinstance(original_text, str):
            raise FieldError(
                f'original_text should be str, not {type(original_text)}'
            )
        self.subject_node = BNode()
        self.original_text = original_text
        self._subfields = [
            ('original_text', EDPOPREC.originalText, 'string'),
            ('normalized_text', EDPOPREC.normalizedText, 'string'),
            ('unknown', EDPOPREC.unknown, 'boolean'),
            ('authority_record', EDPOPREC.authorityRecord, 'string'),
        ]

    def set_normalized_text(self, text: Optional[str]):
        """Manually set the normalized text.

        In case of subclasses that support automatic creation of the 
        normalized text, this method will override the automatic version. 
        Give None as an argument to reset the normalized text."""
        self._normalized_text = text

    @property
    def normalized_text(self) -> Optional[str]:
        """Subfield -- a human-readable string representation of the normalized
        field.

        Should be set manually in the basic ``Field`` class with 
        ``set_normalized_text`` or is automatically created in more complex 
        subclasses. Contains ``None`` in case there is no normalization."""
        if self._normalized_text is not None:
            return self._normalized_text
        if callable(self._create_normalized_text):
            text = self._create_normalized_text()
            assert isinstance(text, str)
            return text
        else:
            return None

    def to_graph(self) -> Graph:
        '''Create an ``rdflib`` RDF graph according to the current data.'''
        assert isinstance(self.subject_node, Node)
        graph = Graph()
        graph.add((
            self.subject_node,
            RDF.type,
            self._rdf_class
        ))
        for subfield in self._subfields:
            attrname, propref, datatype = subfield
            value = getattr(self, attrname, None)
            if value is None:
                # self does not have the attribute or the attribute is None;
                # ignore.
                continue
            try:
                typedef = DATATYPES[datatype]
            except ValueError:
                raise FieldError(
                    f"Datatype '{datatype}' was defined in subfield list on "
                    "{self.__class__} but it does not exist"
                )
            else:
                input_type = typedef['input_type']
                if not isinstance(value, input_type):
                    raise FieldError(
                        f"Subfield {attrname} should be of type {str(input_type)} but "
                        "it is {str(type(value))}"
                    )
                else:
                    converter = typedef['converter']
                    converted = converter(value)
                    assert isinstance(converted, Node)
                    graph.add((
                        self.subject_node,
                        propref,
                        converted
                    ))
        return graph

    def __str__(self) -> str:
        if self.normalized_text is not None:
            return self.normalized_text
        else:
            return self.original_text


class LocationField(Field):
    _rdf_class: Node = EDPOPREC.LocationField
    location_type: Optional[URIRef] = None
    LOCALITY = EDPOPREC.locality
    COUNTRY = EDPOPREC.country

    def __init__(self, original_text: str) -> None:
        super().__init__(original_text)
        self._subfields.append(
            ('location_type', EDPOPREC.locationType, 'uriref')
        )




from rdflib import URIRef
import requests
from typing import List, Dict, Optional

from edpop_explorer import (
    Reader, Record, ReaderError, BiographicalRecord, Field
)


class SBTIReader(Reader):
    api_url = 'https://data.cerl.org/sbti/_search'
    api_by_id_base_url = 'https://data.cerl.org/sbti/'
    link_base_url = 'https://data.cerl.org/sbti/'
    fetching_exhausted: bool = False
    additional_params: Optional[Dict[str, str]] = None
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/sbti'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/sbti/"
    DEFAULT_RECORDS_PER_PAGE = 10

    @classmethod
    def _get_name_field(cls, data: dict) -> Optional[Field]:
        field = None
        firstname = data.get("firstname", None)
        name = data.get("name", None)
        if firstname and name:
            field = Field(f"{firstname} {name}")
        elif name:
            field = Field(f"{name}")
        return field

    @classmethod
    def get_by_id(cls, identifier: str) -> BiographicalRecord:
        try:
            response = requests.get(
                cls.api_by_id_base_url + identifier,
                headers={
                    'Accept': 'application/json'
                },
            ).json()
        except requests.exceptions.JSONDecodeError:
            raise ReaderError(f"Item with id {identifier} does not exist.")
        except requests.exceptions.RequestException as err:
            raise ReaderError(f"Error during server request: {err}")
        return cls._convert_record(response)


    @classmethod
    def _convert_record(cls, rawrecord: dict) -> BiographicalRecord:
        record = BiographicalRecord(from_reader=cls)
        record.data = rawrecord
        record.identifier = rawrecord.get('id', None)
        if not record.identifier:
            record.identifier = rawrecord.get('_id', None)
        if record.identifier:
            record.link = cls.link_base_url + record.identifier

        # Add fields
        heading = rawrecord.get("heading", None)
        if heading:
            name_field = cls._get_name_field(heading[0])
            record.name = name_field
        variant_name = rawrecord.get("variantName", None)
        if isinstance(variant_name, list):
            record.variant_names = []
            for name in variant_name:
                field = cls._get_name_field(name)
                if field:
                    record.variant_names.append(field)
        place_of_activity = rawrecord.get("placeOfActitivty", None)  # sic.
        if isinstance(place_of_activity, list):
            record.places_of_activity = []
            for place in place_of_activity:
                name = place.get("name", None)
                if name:
                    field = Field(name)
                    record.places_of_activity.append(field)

        return record

    def _perform_query(self, start_record: int, maximum_records: Optional[int]) -> List[Record]:
        assert isinstance(self.prepared_query, str)
        if maximum_records is None:
            maximum_records = self.DEFAULT_RECORDS_PER_PAGE
        try:
            response = requests.get(
                self.api_url,
                params={
                    'query': self.prepared_query,
                    'from': start_record,
                    'size': maximum_records,
                    'mode': 'default',
                    'sort': 'default'
                },
                headers={
                    'Accept': 'application/json'
                }
            ).json()
        except (
            requests.exceptions.RequestException
        ) as err:
            raise ReaderError('Error during server request: ' + str(err))

        # TODO: check for error responses
        try:
            if response['hits'] is None:
                self.number_of_results = 0
            else:
                self.number_of_results = response['hits']['value']
        except KeyError:
            raise ReaderError('Number of hits not given in server response')

        if 'rows' not in response:
            # There are no rows in the response, so stop here
            return []

        records: List[Record] = []
        for rawrecord in response['rows']:
            record = self._convert_record(rawrecord)
            records.append(record)

        return records

    @classmethod
    def transform_query(cls, query) -> str:
        # No transformation needed
        return query

    def fetch_range(self, range_to_fetch: range) -> range:
        if self.prepared_query is None:
            raise ReaderError('First call prepare_query')
        if self.fetching_exhausted:
            return range(0)
        start_record = range_to_fetch.start
        number_to_fetch = range_to_fetch.stop - start_record
        results = self._perform_query(start_record, number_to_fetch)
        for i, result in enumerate(results):
            self.records[i] = result
        return range(start_record, start_record + len(results))




from pathlib import Path
import sqlite3
from rdflib import URIRef
import requests
from appdirs import AppDirs
from typing import Optional

from edpop_explorer import (
    Reader, BibliographicalRecord, ReaderError, Field, BIBLIOGRAPHICAL
)
from edpop_explorer.reader import GetByIdBasedOnQueryMixin
from edpop_explorer.sql import SQLPreparedQuery


class FBTEEReader(GetByIdBasedOnQueryMixin, Reader):
    DATABASE_URL = 'https://dhstatic.hum.uu.nl/edpop/cl.sqlite3'
    DATABASE_LICENSE = 'https://dhstatic.hum.uu.nl/edpop/LICENSE.txt'
    FBTEE_LINK = 'http://fbtee.uws.edu.au/stn/interface/browse.php?t=book&' \
        'id={}'
    READERTYPE = BIBLIOGRAPHICAL
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/fbtee'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/fbtee/"
    prepared_query: Optional[SQLPreparedQuery] = None
    FETCH_ALL_AT_ONCE = True

    def __init__(self):
        super().__init__()
        self.database_file = Path(
            AppDirs('edpop-explorer', 'cdh').user_data_dir
        ) / 'cl.sqlite3'

    def prepare_data(self):
        if not self.database_file.exists():
            self._download_database()
        self.con = sqlite3.connect(str(self.database_file))

    def _download_database(self):
        print('Downloading database...')
        response = requests.get(self.DATABASE_URL)
        if response.ok:
            try:
                self.database_file.parent.mkdir(exist_ok=True, parents=True)
                with open(self.database_file, 'wb') as f:
                    f.write(response.content)
            except OSError as err:
                raise ReaderError(
                    f'Error writing database file to disk: {err}'
                )
        else:
            raise ReaderError(
                f'Error downloading database file from {self.DATABASE_URL}'
            )
        print(f'Successfully saved database to {self.database_file}.')
        print(f'See license: {self.DATABASE_LICENSE}')

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> SQLPreparedQuery:
        return SQLPreparedQuery(
            where_statement="WHERE book_code = ?",
            arguments=[identifier]
        )

    @classmethod
    def transform_query(cls, query: str) -> SQLPreparedQuery:
        return SQLPreparedQuery(
            where_statement='WHERE full_book_title LIKE ?',
            arguments=[f'%{query}%']
        )

    @classmethod
    def _add_fields(cls, record: BibliographicalRecord) -> None:
        assert isinstance(record.data, dict)
        record.title = Field(record.data['full_book_title'])
        if record.data['languages']:
            languages = record.data['languages'].split(sep=', ')
            record.languages = [Field(x) for x in languages]
        pages = record.data['pages']
        if pages:
            record.extent = Field(pages)
        place = record.data['stated_publication_places']
        if place:
            record.place_of_publication = Field(place)
        year = record.data['stated_publication_years']
        if year:
            record.dating = Field(year)
        publisher = record.data['stated_publishers']
        if publisher:
            record.publisher_or_printer = Field(publisher)
        record.contributors = []
        for author in record.data['authors']:
            # author is tuple of author code and author name
            record.contributors.append(Field(author[1]))

    def fetch_range(self, range_to_fetch: range) -> range:
        # This method always fetches all data at once. This could be avoided,
        # but it is inexpensive because the data is locally available and
        # the dataset is small.
        self.prepare_data()
        if not self.prepared_query:
            raise ReaderError('First call prepare_query method')
        if self.fetching_exhausted:
            return range(0)
        cur = self.con.cursor()
        columns = [x[1] for x in cur.execute('PRAGMA table_info(books)')]
        res = cur.execute(
            'SELECT B.*, BA.author_code, A.author_name FROM books B '
            'LEFT OUTER JOIN books_authors BA on B.book_code=BA.book_code '
            'JOIN authors A on BA.author_code=A.author_code '
            f'{self.prepared_query.where_statement} '
            'ORDER BY B.book_code',
            self.prepared_query.arguments
        )
        last_book_code = ''
        i = -1
        for row in res:
            # Since we are joining with another table, a book may be repeated,
            # so check if this is a new item
            book_code: str = row[columns.index('book_code')]
            if last_book_code != book_code:
                # We have a new book, so update i
                i += 1
                record = BibliographicalRecord(self.__class__)
                record.data = {}
                for j in range(len(columns)):
                    record.data[columns[j]] = row[j]
                record.identifier = book_code
                record.link = self.FBTEE_LINK.format(book_code)
                record.data['authors'] = []
                self.records[i] = record
                last_book_code = book_code
            # Add author_code and author_name to the last record
            assert len(self.records) > 0
            author_code = row[len(columns)]
            author_name = row[len(columns) + 1]
            assert isinstance(self.records[i].data, dict)
            self.records[i].data['authors'].append((author_code, author_name))
        for record in self.records:
            self._add_fields(record)
        self.number_of_results = len(self.records)
        return range(0, len(self.records))



import sruthi
import requests
from abc import abstractmethod
from typing import List, Optional

from edpop_explorer import Reader, Record, ReaderError
from edpop_explorer.reader import GetByIdBasedOnQueryMixin


class SRUReader(GetByIdBasedOnQueryMixin, Reader):
    '''Subclass of ``Reader`` that adds basic SRU functionality
    using the ``sruthi`` library.

    This class is still abstract and subclasses should implement
    the ``transform_query()`` and ``_convert_record()`` methods
    and set the attributes ``sru_url`` and ``sru_version``.
    
    The ``_prepare_get_by_id_query()`` method by default returns
    the transformed version of the identifier as a query, which
    normally works, but this may be optimised by overriding it.

    .. automethod:: _convert_record'''
    sru_url: str
    '''URL of the SRU API.'''
    sru_version: str
    '''Version of the SRU protocol. Can be '1.1' or '1.2'.'''
    query: Optional[str] = None
    session: requests.Session
    '''The ``Session`` object of the ``requests`` library.'''

    def __init__(self):
        # Set a session to allow reuse of HTTP sessions and to set additional
        # parameters and settings, which some SRU APIs require -
        # see https://github.com/metaodi/sruthi#custom-parameters-and-settings
        super().__init__()
        self.session = requests.Session()

    @classmethod
    @abstractmethod
    def transform_query(cls, query: str) -> str:
        pass

    @classmethod
    @abstractmethod
    def _convert_record(cls, sruthirecord: dict) -> Record:
        '''Convert the output of ``sruthi`` into an instance of
        (a subclass of) ``Record``.'''
        pass

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> str:
        return cls.transform_query(identifier)

    def _perform_query(self, start_record: int, maximum_records: Optional[int]) -> List[Record]:
        if maximum_records is None:
            maximum_records = self.DEFAULT_RECORDS_PER_PAGE
        try:
            response = sruthi.searchretrieve(
                self.sru_url,
                self.prepared_query,
                start_record=start_record,
                maximum_records=maximum_records,
                sru_version=self.sru_version,
                session=self.session
            )
        except (
            sruthi.errors.SruError
        ) as err:
            raise ReaderError('Server returned error: ' + str(err))

        self.number_of_results = response.count

        records: List[Record] = []
        for sruthirecord in response[0:maximum_records]:
            records.append(self._convert_record(sruthirecord))

        return records

    def prepare_query(self, query) -> None:
        self.prepared_query = self.transform_query(query)

    def fetch_range(self, range_to_fetch: range) -> range:
        # SRU provides paged retrieve using a start record (that starts at
        # 1, while we start at 0) and a maximum number of results.
        if self.fetching_exhausted:
            return range(0, 0)
        if self.prepared_query is None:
            raise ReaderError('First call prepare_query')
        start_number = range_to_fetch.start
        start_number_sru = start_number + 1  # SRU starts at 1
        records_to_fetch = range_to_fetch.stop - range_to_fetch.start
        results = self._perform_query(start_number_sru, records_to_fetch)
        for i, result in enumerate(results):
            self.records[i + start_number] = result
        return range(start_number, start_number + len(results))



from typing import Optional, List
from rdflib import URIRef
from edpop_explorer import SRUReader, BibliographicalRecord, BIBLIOGRAPHICAL
from edpop_explorer import Field


class KBReader(SRUReader):
    sru_url = 'http://jsru.kb.nl/sru'
    sru_version = '1.2'
    KB_LINK = 'https://webggc.oclc.org/cbs/DB=2.37/PPN?PPN={}'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/kb'
    )
    READERTYPE = BIBLIOGRAPHICAL
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/kb/"

    def __init__(self):
        super().__init__()
        # The KB SRU requires 'x-collection' as an additional GET parameter
        self.session.params = {
            'x-collection': 'GGC'
        }

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query

    def _find_ppn(self, data: dict):
        """Try to find the PPN given the data that comes from the SRU server;
        return None if PPN cannot be found"""
        # This seems to work fine; not thoroughly tested.
        oai_pmh_identifier = data.get('OaiPmhIdentifier', None)
        if not isinstance(oai_pmh_identifier, str):
            return None
        PREFIX = 'GGC:AC:'
        if oai_pmh_identifier and oai_pmh_identifier.startswith(PREFIX):
            return oai_pmh_identifier[len(PREFIX):]
        return None

    def _convert_record(self, sruthirecord: dict) -> BibliographicalRecord:
        record = BibliographicalRecord(from_reader=type(self))
        record.data = sruthirecord
        record.identifier = self._find_ppn(record.data)
        if record.identifier:
            # Also here: it seems to work, but there may be records where
            # it doesn't work...
            # NOTE: there is often a URL in the `identifier' field as well,
            # but not always, and it uses a `resolver', which is slower.
            # But if we find records for which no link can be found this may
            # be an alternative.
            record.link = self.KB_LINK.format(record.identifier)
        record.title = self._get_title(sruthirecord)
        record.languages = self._get_languages(sruthirecord)
        # TODO: add the other fields
        return record
    
    def _get_title(self, data) -> Optional[Field]:
        if 'title' in data:
            title = data['title']
            if isinstance(title, list):
                # Title contains a list of strings if it consists of multiple
                # parts
                return Field(' : '.join(title))
            else:
                return Field(title)
        else:
            return None

    def _get_languages(self, data) -> Optional[List[Field]]:
        # The 'language' field contains a list of languages, where every
        # language is repeated multiple times in different languages.
        # One of them is always a three-letter language code, so only
        # pass on these. NB: there is a possibility that not all entries
        # consisting of three characters are language codes.
        if 'language' not in data:
            return []
        return [
            Field(x) for x in data['language']
            if isinstance(x, str) and len(x) == 3
        ]



from dataclasses import dataclass, field as dataclass_field
from typing import Dict, List, Optional
import csv
from pathlib import Path
from abc import abstractmethod

from edpop_explorer import (
    BibliographicalRecord, RawData, SRUReader, Field, BIBLIOGRAPHICAL
)


READABLE_FIELDS_FILE = Path(__file__).parent / 'M21_fields.csv'
translation_dictionary: Dict[str, str] = {}
with open(READABLE_FIELDS_FILE) as dictionary_file:
    reader = csv.DictReader(dictionary_file)
    for row in reader:
        translation_dictionary[row['Tag number']] = \
            row[' Tag description'].strip()


@dataclass
class Marc21Field:
    """Python representation of a single field in a Marc21 record"""
    fieldnumber: str
    indicator1: str
    indicator2: str
    subfields: Dict[str, str] = dataclass_field(default_factory=dict)
    description: Optional[str] = None

    def __str__(self):
        '''
        Return the usual marc21 representation
        '''
        sf = []
        ind1 = self.indicator1 if self.indicator1.rstrip() != '' else '#'
        ind2 = self.indicator1 if self.indicator2.rstrip() != '' else '#'
        description = ' ({})'.format(self.description) \
            if self.description else ''
        for subfield in self.subfields:
            sf.append('$${} {}'.format(subfield, self.subfields[subfield]))
        return '{}{}: {} {} {}'.format(
            self.fieldnumber,
            description,
            ind1,
            ind2,
            '  '.join(sf)
        )


@dataclass
class Marc21Data(RawData):
    """Python representation of the data inside a Marc21 record"""
    # We use a list for the fields and not a dictionary because they may
    # appear more than once
    fields: List[Marc21Field] = dataclass_field(default_factory=list)
    controlfields: Dict[str, str] = dataclass_field(default_factory=dict)
    raw: dict = dataclass_field(default_factory=dict)

    def get_first_field(self, fieldnumber: str) -> Optional[Marc21Field]:
        '''Return the first occurance of a field with a given field number.
        May be useful for fields that appear only once, such as 245.
        Return None if field is not found.'''
        for field in self.fields:
            if field.fieldnumber == fieldnumber:
                return field
        return None

    def get_first_subfield(self, fieldnumber: str, subfield: str) -> Optional[str]:
        '''Return the requested subfield of the first occurance of a field with
        the given field number. Return None if field is not found or if the
        subfield is not present on the first occurance of the field.'''
        field = self.get_first_field(fieldnumber)
        if field is not None:
            return field.subfields.get(subfield, None)
        else:
            return None

    def get_fields(self, fieldnumber: str) -> List[Marc21Field]:
        '''Return a list of fields with a given field number. May return an
        empty list if field does not occur.'''
        returned_fields: List[Marc21Field] = []
        for field in self.fields:
            if field.fieldnumber == fieldnumber:
                returned_fields.append(field)
        return returned_fields

    def get_all_subfields(self, fieldnumber: str, subfield: str) -> List[str]:
        '''Return a list of subfields that matches the requested field number
        and subfield. May return an empty list if the field and subfield do not
        occur.'''
        fields = self.get_fields(fieldnumber)
        returned_subfields: List[str] = []
        for field in fields:
            if subfield in field.subfields:
                returned_subfields.append(field.subfields[subfield])
        return returned_subfields

    def to_dict(self) -> dict:
        return self.raw


class Marc21DataMixin():
    """A mixin that adds a ``data`` attribute to a Record class to contain
    an instance of ``Marc21Data``.
    """
    data: Optional[Marc21Data] = None

    def show_record(self) -> str:
        if self.data is None:
            return "(no data)"
        field_strings = []
        for field in self.data.fields:
            field_strings.append(str(field))
        return '\n'.join(field_strings)

class SRUMarc21Reader(SRUReader):
    '''Subclass of ``SRUReader`` that adds Marc21 functionality.

    This class is still abstract and to create concrete readers
    the ``_get_link()``, ``_get_identifier()`` 
    and ``_convert_record`` methods should be implemented.

    .. automethod:: _convert_record
    .. automethod:: _get_link
    .. automethod:: _get_identifier'''
    marcxchange_prefix: str = ''

    @classmethod
    def _get_subfields(cls, sruthifield) -> list:
        # If there is only one subfield, sruthi puts it directly in
        # a dict, otherwise it uses a list of dicts. Make sure that
        # we always have a list.
        subfielddata = sruthifield[f'{cls.marcxchange_prefix}subfield']
        if isinstance(subfielddata, dict):
            sruthisubfields = [subfielddata]
        else:
            sruthisubfields = subfielddata
        assert isinstance(sruthisubfields, list)
        return sruthisubfields

    @classmethod
    def _convert_to_marc21data(cls, sruthirecord: dict) -> Marc21Data:
        data = Marc21Data()
        data.raw = sruthirecord
        # marcxml (marc21 in xml) consists of a controlfield and a datafield.
        # The controlfield and the datafield contain multiple fields.
        # The controlfield consists of simple pairs of tags (field numbers)
        # and texts (field values).
        for sruthicontrolfield in \
                sruthirecord[f'{cls.marcxchange_prefix}controlfield']:
            tag = sruthicontrolfield['tag']
            text = sruthicontrolfield['text']
            data.controlfields[tag] = text
        # The datafield is more complex; these fields also have two indicators,
        # one-digit numbers that carry special meanings, and multiple subfields
        # that each have a one-character code.
        for sruthifield in sruthirecord[f'{cls.marcxchange_prefix}datafield']:
            fieldnumber = sruthifield['tag']
            field = Marc21Field(
                fieldnumber=fieldnumber,
                indicator1=sruthifield['ind1'],
                indicator2=sruthifield['ind2'],
                subfields={}
            )
            # The translation_dictionary contains descriptions for a number
            # of important fields. Include them so that the user can more
            # easily understand the record.
            if fieldnumber in translation_dictionary:
                field.description = translation_dictionary[fieldnumber]
            sruthisubfields = cls._get_subfields(sruthifield)

            for sruthisubfield in sruthisubfields:
                field.subfields[sruthisubfield['code']] = \
                    sruthisubfield['text']
            data.fields.append(field)
        return data
    
    @classmethod
    @abstractmethod
    def _get_link(cls, data: Marc21Data) -> Optional[str]:
        '''Get a public URL according to the Marc21 data or ``None`` if it
        is not available.'''
        pass

    @classmethod
    @abstractmethod
    def _get_identifier(cls, data: Marc21Data) -> Optional[str]:
        '''Get the unique identifier from the Marc21 data or ``None`` if it
        is not available.'''
        pass


class Marc21BibliographicalRecord(Marc21DataMixin, BibliographicalRecord):
    '''A combination of ``BibliographicalRecord`` and ``Marc21DataMixin``.'''
    pass


class SRUMarc21BibliographicalReader(SRUMarc21Reader):
    '''Subclass of ``SRUMarc21Reader`` that adds functionality to create
    instances of ``BibliographicRecord``.

    This subclass assumes that the Marc21 data is according to the standard
    format of Marc21 for bibliographical data. See:
    https://www.loc.gov/marc/bibliographic/
    '''
    _title_field_subfield = ('245', 'a')
    _alternative_title_field_subfield = ('246', 'a')
    _publisher_field_subfield = ('264', 'b')
    _language_field_subfield = ('041', 'a')
    _place_field_subfield = ('264', 'a')
    _dating_field_subfield = ('264', 'c')
    _extent_field_subfield = ('300', 'a')
    _physical_description_field_subfield = ('300', 'b')
    _size_field_subfield = ('300', 'c')

    records: List[Marc21BibliographicalRecord]
    READERTYPE = BIBLIOGRAPHICAL
    
    @classmethod
    def _convert_record(cls, sruthirecord: dict) -> Marc21BibliographicalRecord:
        record = Marc21BibliographicalRecord(from_reader=cls)
        data = cls._convert_to_marc21data(sruthirecord)
        record.data = data
        record.link = cls._get_link(data)
        record.identifier = cls._get_identifier(data)
        # NOTE: it is probably better to outfactor the following logic to
        # other class methods, to offer more flexibility and because this
        # will become more complex as we add normalization.
        title = data.get_first_subfield(*cls._title_field_subfield)
        if title:
            record.title = Field(title)
        alternative_title = data.get_first_subfield(
            *cls._alternative_title_field_subfield
        )
        if alternative_title:
            record.alternative_title = Field(alternative_title)
        publisher = data.get_first_subfield(*cls._publisher_field_subfield)
        if publisher:
            record.publisher_or_printer = Field(publisher)
        place = data.get_first_subfield(*cls._place_field_subfield)
        if place:
            record.place_of_publication = Field(place)
        language = data.get_first_subfield(*cls._language_field_subfield)
        # TODO: look up if this field is repeatable - if so support multiple
        # languages
        if language:
            record.languages = [Field(language)]
        dating = data.get_first_subfield(*cls._dating_field_subfield)
        if dating:
            record.dating = Field(dating)
        extent = data.get_first_subfield(*cls._extent_field_subfield)
        if extent:
            record.extent = Field(extent)
        physical_description = data.get_first_subfield(
            *cls._physical_description_field_subfield
        )
        if physical_description:
            record.physical_description = Field(physical_description)
        size = data.get_first_subfield(*cls._size_field_subfield)
        if size:
            record.size = Field(size)

        # Add the contributors
        record.contributors = cls._get_contributors(data)

        return record

    @classmethod
    def _get_contributors(cls, data: Marc21Data) -> List[Field]:
        contributors: List[Field] = []
        contributor_fields = data.get_fields('100')
        for field in contributor_fields:
            name = field.subfields.get('a')
            if name:
                contributors.append(Field(name))
        return contributors




from typing import Optional, Type
from rdflib import Graph
import json
from SPARQLWrapper import SPARQLWrapper, SPARQLExceptions, JSON as JSONFormat
from abc import abstractmethod

from typing_extensions import override

from edpop_explorer import (
    Reader, Record, BibliographicalRecord, ReaderError, RecordError,
    LazyRecordMixin
)

PREFIXES = {
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'rdfs': 'http://www.w3.org/2000/01/rdf-schema#',
    'schema': 'http://schema.org/',
    'owl': 'http://www.w3.org/2002/07/owl#',
}

PREFIXES_REVERSE_REPLACEMENT_TABLE = {
    PREFIXES[key]: (key + ':') for key in PREFIXES
}

PREFIX_DEFINITIONS = '\n'.join([
    f'prefix {key}: <{PREFIXES[key]}>' for key in PREFIXES
])

prepare_listing_query = (PREFIX_DEFINITIONS + """
select ?s ?name where
{{
  ?s ?p ?o .
  ?s {name_predicate} ?name .
  {filter}
  FILTER (regex(?o, "{query}","i"))
}}
order by ?s
""").format

prepare_lookup_query = """
prefix schema: <http://schema.org/>
select ?p ?o
{{
    <{identifier}> ?p ?o
}}
""".format


def replace_fqu_with_prefixed_uris(inputstring: str) -> str:
    '''Replace fully qualified URIs to prefixed URIs if they occur in
    the prefix table in the prefixes attribute'''
    for key in PREFIXES_REVERSE_REPLACEMENT_TABLE:
        inputstring = inputstring.replace(
            key, PREFIXES_REVERSE_REPLACEMENT_TABLE[key], 1
        )
    return inputstring


class RDFRecordMixin(LazyRecordMixin):
    '''Mixin that adds lazy RDF fetching functionality to a Record.'''
    identifier: Optional[str] = None
    fetched: bool = False
    data: Optional[dict] = None
    original_graph: Optional[Graph] = None
    from_reader: Type["SparqlReader"]

    def fetch(self) -> None:
        # TODO: at the moment this mixin only supports fetching RDF data
        # if the IRI in the identifier attribute is available via HTTP
        # as data that rdflib can process. We might need to support
        # IRIs that can only be accessed via an endpoint as well.
        if not self.identifier:
            raise RecordError(
                'identifier (subject IRI) has not been set'
            )
        if self.fetched:
            return
        try:
            self.original_graph = Graph()
            self.original_graph.parse(self.identifier)
        except Exception as err:
            # URLLib does not catch errors of underlying libraries, hence
            # the use of except Exception
            raise RecordError(
                f"Error while loading record's contents from IRI "
                f"{self.identifier}: {err}"
            )
        # Convert to JSON for raw data attribute
        self.data = json.loads(
            self.original_graph.serialize(format="json-ld")
        )
        # Call Reader's data conversion method to fill the record's Fields
        assert isinstance(self, Record)
        self.from_reader._convert_record(self.original_graph, self)

        self.fetched = True


class BibliographicalRDFRecord(RDFRecordMixin, BibliographicalRecord):
    pass


class SparqlReader(Reader):
    endpoint: str
    name_predicate: str
    filter: Optional[str] = None
    prepared_query: Optional[str]
    FETCH_ALL_AT_ONCE = True

    @classmethod
    @override
    def transform_query(cls, query: str):
        return prepare_listing_query(
            name_predicate=cls.name_predicate,
            filter=cls.filter,
            query=query
        )

    @classmethod
    @override
    def get_by_id(cls, identifier: str) -> Record:
        return cls._create_lazy_record(identifier)

    @override
    def fetch_range(self, range_to_fetch: range) -> range:
        # Fetch all records at one, because this is an expensive operation.
        if not self.prepared_query:
            raise ReaderError('First call prepare_query method')
        if self.fetching_exhausted:
            return range(0, 0)
        wrapper = SPARQLWrapper(self.endpoint)
        wrapper.setReturnFormat(JSONFormat)
        wrapper.setQuery(self.prepared_query)
        try:
            response = wrapper.queryAndConvert()
        except SPARQLExceptions.QueryBadFormed as err:
            raise ReaderError(
                'Malformed SPARQL query: {}'.format(err)
            )
        assert isinstance(response, dict)
        results = response['results']['bindings']
        self.records = {}
        self.number_of_results = len(results)
        for i, result in enumerate(results):
            iri = result['s']['value']
            name = result['name']['value']
            self.records[i] = self._create_lazy_record(iri, name)
        return range(0, self.number_of_results)

    @classmethod
    @abstractmethod
    def convert_record(cls, graph: Graph, record: Record) -> None:
        '''Convert data from an RDF graph to Fields in a Record. The 
        Record is changed in-place.'''
        pass

    @classmethod
    @abstractmethod
    def _create_lazy_record(cls, iri: str, name: Optional[str]=None) -> Record:
        """Create a Record/LazyRecordMixin record object.

        This is the lazy record that is created after running the SPARQL
        query. The record initially only gets the IRI attached as well
        as a single property (``name``, used for quick identification), while
        the rest will be loaded when Record.fetch() is called.
        """
        pass



from unittest.mock import patch
import json
from pathlib import Path
from typing import Optional

from edpop_explorer import SRUMarc21BibliographicalReader, Marc21Data


TESTDATA = json.load(open(Path(__file__).parent / 'TESTDATA', 'r'))


class MockReader(SRUMarc21BibliographicalReader):
    @classmethod
    def transform_query(cls, query: str) -> str:
        return query

    @classmethod
    def _get_link(cls, data: Marc21Data) -> Optional[str]:
        return "https://www.example.com"

    @classmethod
    def _get_identifier(cls, data: Marc21Data) -> Optional[str]:
        return 'id'


class TestSRUMarc21BibliographicalReader:
    @patch('edpop_explorer.srureader.sruthi')
    def test_fetch(self, mock_sruthi):
        mock_sruthi.searchretrieve.return_value = TESTDATA
        reader = MockReader()
        reader.sru_url = ''
        reader.sru_version = '1.1'
        reader.prepare_query('testquery')
        reader.fetch()
        results = reader.records
        # Field with multiple subfields
        data = results[0].data
        assert data is not None
        firstfield = data.get_first_field('245')
        assert firstfield is not None
        assert firstfield.subfields['a'] == \
            'Aeschylus: Eumenides.'
        # Field with a single subfield
        firstfield = data.get_first_field('650')
        assert firstfield is not None
        assert firstfield.subfields['a'] == \
            'Aeschylus Eumenides.'
        # Field's description
        assert firstfield.description == \
            'Subject Added Entry - Topical Term'
        # Field that occurs multiple times
        assert len(data.get_fields('500')) == 5
        # Control field
        assert data.controlfields['007'] == 'tu'



from pytest import fixture, raises
from rdflib import Literal, RDF
from rdflib.term import Node

from edpop_explorer import Field, FieldError, LocationField
from edpop_explorer import EDPOPREC


@fixture
def basic_field() -> Field:
    return Field('Dit is een boektitel')


@fixture
def basic_location_field() -> LocationField:
    field = LocationField('Voorschoten')
    return field


class TestField:
    def test_init(self, basic_field: Field):
        assert basic_field.original_text == 'Dit is een boektitel'
        assert isinstance(basic_field.subject_node, Node)

    def test_to_graph(self, basic_field: Field):
        graph = basic_field.to_graph()
        assert (basic_field.subject_node, RDF.type, EDPOPREC.Field) in graph
        # Test simple string
        assert (
            basic_field.subject_node,
            EDPOPREC.originalText,
            Literal(basic_field.original_text)
        ) in graph
        # Test string from property
        basic_field.set_normalized_text('normalized')
        graph = basic_field.to_graph()
        assert (
            basic_field.subject_node,
            EDPOPREC.normalizedText,
            Literal(basic_field.normalized_text)
        ) in graph
        # Test boolean
        basic_field.unknown = True
        graph = basic_field.to_graph()
        assert (
            basic_field.subject_node,
            EDPOPREC.unknown,
            Literal(True)
        ) in graph
        # Invalid type on object should give exception
        basic_field.unknown = 'other value'  # type: ignore
        with raises(FieldError):
            basic_field.to_graph()
        # Nonexisting datatype defined in class on SUBFIELDS should give
        # exception
        basic_field._subfields = basic_field._subfields.copy()
        basic_field._subfields.append(
            ('other', EDPOPREC.other, 'othertype')
        )
        basic_field.other = 'text'  # type: ignore
        with raises(FieldError):
            basic_field.to_graph()

    def test_normalized_text(self, basic_field: Field):
        # If nothing is set, this should be None
        assert basic_field.normalized_text is None
        # Set normalized text by hand
        text = 'normalized'
        basic_field.set_normalized_text(text)
        assert basic_field.normalized_text == text
        # Now test a class with automatic normalized text creation

        class ComplexField(Field):
            def _create_normalized_text(self):
                return self.original_text.capitalize()
        title = 'title'
        complex_field = ComplexField(title)
        assert complex_field.normalized_text == title.capitalize()
        # A manual normalized text should override this
        complex_field.set_normalized_text(text)
        assert complex_field.normalized_text == text


class TestLocationField:
    def test_basic_form(self, basic_location_field: LocationField):
        field = basic_location_field
        graph = field.to_graph()
        assert (
            field.subject_node,
            EDPOPREC.locationType,
            None
        ) not in graph

    def test_location_type(self, basic_location_field: LocationField):
        field = basic_location_field
        field.location_type = LocationField.LOCALITY
        graph = field.to_graph()
        assert (
            field.subject_node,
            EDPOPREC.locationType,
            EDPOPREC.locality
        ) in graph



from typing import Optional

from rdflib import URIRef

from edpop_explorer import SRUMarc21BibliographicalReader, Marc21Data


class VDCommonMixin():
    LINK_FORMAT: str

    @classmethod
    def _get_identifier(cls, data: Marc21Data) -> Optional[str]:
        field024 = data.get_first_field('024')
        if field024:
            return field024.subfields.get('a', None)
        else:
            return None

    @classmethod
    def _get_link(cls, record: Marc21Data) -> Optional[str]:
        identifier = cls._get_identifier(record)
        if identifier:
            return cls.LINK_FORMAT.format(identifier).replace(' ', '+')


class VD16Reader(VDCommonMixin, SRUMarc21BibliographicalReader):
    sru_url = 'http://bvbr.bib-bvb.de:5661/bvb01sru'
    sru_version = '1.1'
    LINK_FORMAT = 'http://gateway-bayern.de/{}'  # Spaces should be replaced by +
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/vd16'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/vd16/"

    @classmethod
    def transform_query(cls, query: str) -> str:
        # This SRU URL combines multiple databases, so make sure only VD16 is
        # queried
        return 'VD16 and ({})'.format(query)


class VD17Reader(VDCommonMixin, SRUMarc21BibliographicalReader):
    sru_url = 'http://sru.k10plus.de/vd17'
    sru_version = '1.1'
    LINK_FORMAT = \
        'https://kxp.k10plus.de/DB=1.28/CMD?ACT=SRCHA&IKT=8079&TRM=%27{}%27'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/vd17'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/vd17/"

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query


class VD18Reader(VDCommonMixin, SRUMarc21BibliographicalReader):
    sru_url = 'http://sru.k10plus.de/vd18'
    sru_version = '1.1'
    LINK_FORMAT = 'https://kxp.k10plus.de/DB=1.65/SET=1/TTL=1/CMD?ACT=SRCHA&' \
        'IKT=1016&SRT=YOP&TRM={}&ADI_MAT=B&MATCFILTER=Y&MATCSET=Y&ADI_MAT=T&' \
        'REC=*'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/vd18'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/vd18/"

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query

    @classmethod
    def _get_identifier(cls, record: Marc21Data):
        # The record id is in field 024 for which subfield 2 is vd18. There
        # may be more than one occurance of field 024.
        fields024 = record.get_fields('024')
        for field in fields024:
            if '2' in field.subfields and \
                    'a' in field.subfields and \
                    field.subfields['2'] == 'vd18':
                return field.subfields['a'][5:]
        return None


class VDLiedReader(VDCommonMixin, SRUMarc21BibliographicalReader):
    sru_url = 'http://sru.k10plus.de/vdlied'
    sru_version = '1.1'
    LINK_FORMAT = 'https://gso.gbv.de/DB=1.60/PPNSET?PPN={}'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/vdlied'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/vdlied/"

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query

    @classmethod
    def _get_identifier(cls, record: Marc21Data) -> Optional[str]:
        return record.controlfields.get("001", None)



from pathlib import Path
import sqlite3
from typing import List, Optional, Union
from appdirs import AppDirs
from rdflib import URIRef

from edpop_explorer import (
    Reader, BibliographicalRecord, ReaderError, Field, BIBLIOGRAPHICAL,
    GetByIdBasedOnQueryMixin
)
from edpop_explorer.sql import SQLPreparedQuery


class USTCReader(GetByIdBasedOnQueryMixin, Reader):
    DATABASE_FILENAME = 'ustc.sqlite3'
    USTC_LINK = 'https://www.ustc.ac.uk/editions/{}'
    READERTYPE = BIBLIOGRAPHICAL
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/ustc'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/ustc/"
    prepared_query: Optional[SQLPreparedQuery] = None
    FETCH_ALL_AT_ONCE = True

    def __init__(self):
        super().__init__()
        self.database_file = Path(
            AppDirs('edpop-explorer', 'cdh').user_data_dir
        ) / self.DATABASE_FILENAME

    def prepare_data(self):
        if not self.database_file.exists():
            # Find database dir with .resolve() because on Windows it is
            # some sort of hidden symlink if Python was installed using
            # the Windows Store...
            db_dir = self.database_file.parent.resolve()
            print(f'USTC database not found. Please obtain the file '
                  f'{self.DATABASE_FILENAME} from the project team and add it '
                  f'to the following directory: {db_dir}')
            raise ReaderError('Database file not found')
        self.con = sqlite3.connect(str(self.database_file))

    @classmethod
    def transform_query(cls, query: str) -> SQLPreparedQuery:
        where_statement = ( 
            'WHERE E.std_title LIKE ? '
            'OR E.author_name_1 LIKE ? '
            'OR E.author_name_2 LIKE ? '
            'OR E.author_name_3 LIKE ? '
            'OR E.author_name_4 LIKE ? '
            'OR E.author_name_5 LIKE ? '
            'OR E.author_name_6 LIKE ? '
            'OR E.author_name_7 LIKE ? '
            'OR E.author_name_8 LIKE ? '
        )
        like_argument = '%' + query + '%'
        arguments: List[Union[str, int]] = [like_argument for _ in range(9)]
        return SQLPreparedQuery(where_statement, arguments)

    @classmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> SQLPreparedQuery:
        try:
            identifier_int = int(identifier)
        except ValueError:
            raise ReaderError(f"Identifier {identifier} is not an integer")
        return SQLPreparedQuery(
            where_statement="WHERE E.sn = ?",
            arguments=[identifier_int]
        )

    def fetch_range(self, range_to_fetch: range) -> range:
        self.prepare_data()

        # This method fetches all records immediately, because the data is
        # locally stored.

        if not self.prepared_query:
            raise ReaderError('No query has been set')
        if self.fetching_exhausted:
            return range(0)

        cur = self.con.cursor()
        columns = [x[1] for x in cur.execute('PRAGMA table_info(editions)')]
        # This kind of query is far from ideal, but the alternative is to
        # implement SQLite full text search which is probably too much work
        # for our current goal (i.e. getting insight in the data structures)
        res = cur.execute(
            'SELECT E.* FROM editions E '
            + self.prepared_query.where_statement
            + ' ORDER BY E.id',
            self.prepared_query.arguments,
        )
        for i, row in enumerate(res):
            data = {}
            for j in range(len(columns)):
                data[columns[j]] = row[j]
            record = self._convert_record(data)
            self.records[i] = record
        self.number_of_results = len(self.records)
        return range(0, len(self.records))

    def _convert_record(self, data: dict) -> BibliographicalRecord:
        record = BibliographicalRecord(from_reader=self.__class__)
        record.data = data
        record.identifier = data['sn']
        record.link = self.USTC_LINK.format(data['sn'])
        record.title = Field(data['std_title'])
        record.contributors = []
        for i in range(8):
            fieldname = f'author_name_{i + 1}'
            if data[fieldname]:
                record.contributors.append(Field(data[fieldname]))
        if data['printer_name_1']:
            # TODO: support for multiple printers
            record.publisher_or_printer = Field(data['printer_name_1'])
        if data['place']:
            record.place_of_publication = Field(data['place'])
        if data['year']:
            record.dating = Field(data['year'])
        record.languages = []
        for i in range(4):
            fieldname = f'language_{i + 1}'
            if data[fieldname]:
                record.languages.append(Field(data[fieldname]))
        if data['pagination']:
            record.extent = Field(data['pagination'])
        return record




"""Test all readers for basic sanity, without invoking methods that could
access the internet or manipulate files."""

import pytest
import warnings
from typing import Type

from rdflib import Graph

from edpop_explorer import Reader, ReaderError
from edpop_explorer.readers import ALL_READERS


@pytest.mark.parametrize("readercls", ALL_READERS)
def test_instantiate(readercls: Type[Reader]):
    # Test if all concrete reader classes can be instantiated: all abstract
    # methods should have been overridden and there should be no errors
    # in the class definition.
    readercls()


@pytest.mark.parametrize("readercls", ALL_READERS)
def test_transform_query(readercls: Type[Reader]):
    readercls.transform_query("testquery")


@pytest.mark.parametrize("readercls", ALL_READERS)
def test_identifier_to_iri(readercls: Type[Reader]):
    # Converting an identifier to an IRI should always be possible
    assert isinstance(readercls.identifier_to_iri("123"), str)


@pytest.mark.parametrize("readercls", ALL_READERS)
def test_iri_to_identifier(readercls: Type[Reader]):
    # This is likely to fail, but if it fails it should raise a ReaderError
    # or derivative
    try:
        readercls.iri_to_identifier("http://example.com/record/1")
    except ReaderError:
        pass


@pytest.mark.parametrize("readercls", ALL_READERS)
def test_catalog_to_graph(readercls: Type[Reader]):
    assert isinstance(readercls.catalog_to_graph(), Graph)


@pytest.mark.parametrize("readercls", ALL_READERS)
@pytest.mark.requests
def test_realrequest(readercls: Type[Reader]):
    reader = readercls()
    reader.prepare_query("gruninger")
    rng = reader.fetch(5)
    assert reader.number_of_results is not None
    assert reader.number_fetched == len(reader.records)
    assert rng == range(0, reader.number_fetched)
    if not reader.fetching_exhausted:
        # Assert that maximum number of results is respected if reader does
        # not fetch all results at once
        assert reader.number_fetched <= 5
    assert reader.number_of_results >= reader.number_fetched
    if reader.number_fetched > 0:
        record = reader.records[0]
        assert record is not None
        record.fetch()
        assert isinstance(record.to_graph(), Graph)
        # Take the IRI and check if searching by IRI gives
        # the same result
        iri = record.iri
        if iri is not None:
            samerecord = reader.get_by_iri(iri)
            assert samerecord.iri == iri
        else:
            warnings.warn(UserWarning(f"Record {record} has empty IRI"))
    # Perform a second fetch
    fetched_before = reader.number_fetched
    rng2 = reader.fetch()  # Do not pass number of results to test that as well
    # If not all records had been fetched already, more records
    # should be available now. Otherwise, nothing should have
    # changed.
    if fetched_before < reader.number_of_results:
        assert reader.number_fetched > fetched_before
        assert rng2.start == fetched_before
        assert rng2.stop == reader.number_fetched
        # Assert that the last record of the previous fetch is not the same as
        # the first record of the current fetch
        record1 = reader.records[fetched_before - 1]
        record2 = reader.records[fetched_before]
        assert record1 is not None and record2 is not None
        if record1.identifier is not None and record1.identifier == record2.identifier:
            # Both records have the same identifier, which could mean that
            # there was a mistake with the offsets. But just give a warning,
            # because there are APIs that (by mistake?) return duplicated
            # records.
            warnings.warn(UserWarning("Last record from first fetch is same as first record from second fetch"))
    else:
        assert reader.number_fetched == fetched_before
        assert rng2 == range(0)



"""Base reader class and strongly related functionality."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, Union, Dict
from rdflib import Graph, RDF, URIRef
from urllib.parse import quote, unquote


from edpop_explorer import (
    EDPOPREC, BIBLIOGRAPHICAL, BIOGRAPHICAL, bind_common_namespaces
)
from .record import Record


@dataclass
class BasePreparedQuery:
    """Empty base dataclass for prepared queries. For prepared queries that
    can be represented by a single string, do not inherit from this class
    but use a simple string instead."""
    pass


PreparedQueryType = Union[str, BasePreparedQuery]


class Reader(ABC):
    """Base reader class (abstract).

    This abstract base class provides a common interface for all readers.
    To use, instantiate a subclass, set a query using the
    ``prepare_query()`` or ``set_query()`` method, call ``fetch()``
    and subsequently ``fetch_next()`` until you have the number
    of results that you want. The attributes ``number_of_results``,
    ``number_fetched`` and ``records`` will be updated after
    fetching.

    To create a concrete reader, make a subclass that implements the
    ``fetch_range()`` and ``transform_query()`` methods
    and set the ``READERTYPE`` and ``CATALOG_URIREF`` attributes.
    ``fetch_range()`` should populate the ``records``, ``number_of_results``,
    ``number_fetched`` and ``range_fetched`` attributes.
    """
    number_of_results: Optional[int] = None
    """The total number of results for the query, or None if fetching
    has not yet started and the number is not yet known."""
    records: Dict[int, Record]
    """The records that have been fetched as instances of
    (a subclass of) ``Record``."""
    prepared_query: Optional[PreparedQueryType] = None
    """A transformed version of the query, available after
    calling ``prepare_query()`` or ``set_query``."""
    READERTYPE: Optional[str] = None
    """The type of the reader, out of ``BIOGRAPHICAL`` and
    ``BIBLIOGRAPHICAL`` (defined in the ``edpop_explorer`` package)."""
    CATALOG_URIREF: Optional[URIRef] = None
    IRI_PREFIX: Optional[str] = None
    """The prefix to use to create an IRI out of a record identifier.
    If an IRI cannot be created with a simple prefix, the 
    `identifier_to_iri` and `iri_to_identifier` methods have to be
    overridden."""
    FETCH_ALL_AT_ONCE = False
    """True if the reader is configured to fetch all records at once,
    even if the user only needs a subset."""
    DEFAULT_RECORDS_PER_PAGE: int = 10
    """The number of records to fetch at a time using the ``fetch()``
    method if not determined by user."""
    _fetch_position: int = 0
    """The index of the record that was fetched last. This is used by
    the ``fetch()`` method to decide where to continue fetching."""

    def __init__(self):
        self.records = {}

    @classmethod
    @abstractmethod
    def transform_query(cls, query: str) -> PreparedQueryType:
        """Return a version of the query that is prepared for use in the
        API.

        This method does not have to be called directly; instead
        ``prepare_query()`` can be used."""
        pass

    def prepare_query(self, query: str) -> None:
        """Prepare a query for use by the reader's API. Updates the
        ``prepared_query`` attribute."""
        self.prepared_query = self.transform_query(query)

    def set_query(self, query: PreparedQueryType) -> None:
        """Set an exact query. Updates the ``prepared_query``
        attribute."""
        self.prepared_query = query

    def adjust_start_record(self, start_number: int) -> None:
        """Skip the given number of first records and start fetching
        afterwards.

        This functionality may be ignored by readers that can only load
        all records at once; generally these are readers that return lazy
        records."""
        self._fetch_position = start_number

    def fetch(
            self, number: Optional[int] = None
    ) -> range:
        """Perform an initial or subsequent query. Most readers fetch
        a limited number of records at once -- this number depends on
        the reader but it may be adjusted using the ``number`` parameter.
        Other readers fetch all records at once and ignore the ``number``
        parameter. After fetching, the records are available in the
        ``records`` attribute and the ``number_of_results`` attribute
        will be available. Returns the range of record indexes that has
        been fetched."""
        if self.fetching_exhausted:
            return range(0)
        if number is None:
            number = self.DEFAULT_RECORDS_PER_PAGE
        resulting_range = self.fetch_range(range(self._fetch_position,
                                           self._fetch_position + number))
        self._fetch_position = resulting_range.stop
        return resulting_range

    @abstractmethod
    def fetch_range(self, range_to_fetch: range) -> range:
        """Fetch a specific range of records. After fetching, the records
        are available in the ``records`` attribute and the
        ``number_of_results`` attribute will be available. If not all records
        of the specified range exist, only the records that exist will be
        fetched.

        :param range_to_fetch: The range of records to fetch. ``step`` values
            of ranges other than 1 are not supported and may be ignored.
        :returns: The range of record indexes that has actually been fetched.
        """
        pass

    def get(self, index: int, allow_fetching: bool = True) -> Record:
        """Get a record with a specific index. If the record is not yet
        available, fetch additional records to make it available.

        :param index: The number of the record to get.
        :param allow_fetching: Allow fetching the record from an external
            source if it was not yet fetched.
        """
        try:
            return self.records[index]
        except KeyError:
            record = None
            # Try to fetch, if it is allowed, and if there is a chance that
            # it is successful (by verifying that index is not out of
            # available range, if known)
            if (allow_fetching and
                    (self.number_of_results is None
                     or self.number_of_results <= index)):
                # Fetch and try again
                self.fetch_range(range(index, index + 1))
                record = self.records.get(index)
            if record is not None:
                return record
            else:
                raise NotFoundError(f"Item with index {index} is not available.")

    @classmethod
    @abstractmethod
    def get_by_id(cls, identifier: str) -> Record:
        """Get a single record by its identifier."""
        pass

    @classmethod
    def get_by_iri(cls, iri: str) -> Record:
        """Get a single records by its IRI."""
        identifier = cls.iri_to_identifier(iri)
        return cls.get_by_id(identifier)

    @classmethod
    def identifier_to_iri(cls, identifier: str) -> str:
        if not isinstance(cls.IRI_PREFIX, str):
            raise ReaderError(
                f"Cannot convert identifier to IRI: {__class__}.IRI_PREFIX "
                "not a string."
            )
        return cls.IRI_PREFIX + quote(identifier)

    @classmethod
    def iri_to_identifier(cls, iri: str) -> str:
        if not isinstance(cls.IRI_PREFIX, str):
            raise ReaderError(
                f"Cannot convert IRI to identifier: {__class__}.IRI_PREFIX "
                "not a string."
            )
        if iri.startswith(cls.IRI_PREFIX):
            return unquote(iri[len(cls.IRI_PREFIX):])
        else:
            raise ReaderError(
                f"Cannot convert IRI {iri} to identifier: IRI does not start "
                "with {cls.IRI_PREFIX}."
            )

    @classmethod
    def catalog_to_graph(cls) -> Graph:
        '''Create an RDF representation of the catalog that this reader
        supports as an instance of EDPOPREC:Catalog.'''
        g = Graph()
        if not cls.CATALOG_URIREF:
            raise ReaderError(
                'Cannot create graph because catalog IRI has not been set. '
                'This should have been done on class level.'
            )

        # Set reader class
        rdfclass = EDPOPREC.Catalog
        if cls.READERTYPE == BIOGRAPHICAL:
            rdfclass = EDPOPREC.BiographicalCatalog
        elif cls.READERTYPE == BIBLIOGRAPHICAL:
            rdfclass = EDPOPREC.BibliographicalCatalog
        g.add((cls.CATALOG_URIREF, RDF.type, rdfclass))

        # Set namespace prefixes
        bind_common_namespaces(g)

        return g

    @property
    def fetching_exhausted(self) -> bool:
        """Return ``True`` if all results have been fetched."""
        return self.fetching_started and self.number_of_results == self.number_fetched

    @property
    def fetching_started(self) -> bool:
        """``True`` if fetching has started, otherwise ``False``. As soon
        as fetching has started, changing the query is not possible anymore."""
        return self.number_of_results is not None

    @property
    def number_fetched(self) -> int:
        """The number of results that has been fetched so far, or 0 if
        no fetch has been performed yet."""
        return len(self.records)

    def generate_identifier(self) -> str:
        """Generate an identifier for this reader that is unique for the
        combination of reader type and prepared query. This identifier can
        be used when the reader has to be reused across sessions by
        pickling and unpickling.

        Note: while the identifier is guaranteed to be unique, there
        is no guarantee that the generated identifier is the same for
        every combination of reader type and prepared query."""
        if self.prepared_query is None:
            raise RuntimeError("A prepared query should be set first")
        # Create identifier based on reader class name and prepared query.
        readertype = self.__class__
        # self.prepared_query is either a string or a dataclass instance,
        # which means that it has a __str__ method that gives a unique
        # string representation of its contents (at least as long as
        # it does not contain a very complex data structure, which should
        # not be the case). For dataclasses, it is not guaranteed
        prepared_query = str(self.prepared_query)
        return f"{readertype} | {prepared_query}"


class GetByIdBasedOnQueryMixin(ABC):
    """Mixin for readers that are based on an API that has no special
    way of retrieving single records -- instead, these readers fetch
    single records using a list query. To use, make sure to override
    the ``_prepare_get_by_id_query`` method, which defines the list
    query that should be used."""

    @classmethod
    def get_by_id(cls, identifier: str) -> Record:
        reader = cls()
        assert isinstance(reader, Reader), \
            "GetByIdBasedOnQueryMixin should be used on Reader subclass"
        reader.set_query(cls._prepare_get_by_id_query(identifier))
        reader.fetch()
        if reader.number_of_results == 0:
            raise NotFoundError("No results returned")
        for record in reader.records.values():
            assert record is not None
            if record.identifier == identifier:
                return record
        # Record with correct ID was not returned in first fetch -
        # give up.
        raise NotFoundError(
            f"Record with identifier {identifier} not present among "
            f"{reader.number_of_results} returned results."
        )

    @classmethod
    @abstractmethod
    def _prepare_get_by_id_query(cls, identifier: str) -> PreparedQueryType:
        pass


class ReaderError(Exception):
    """Generic exception for failures in ``Reader`` class. More specific errors
    derive from this class."""
    pass


class NotFoundError(ReaderError):
    pass



from rdflib import URIRef
from edpop_explorer import Record, SRUReader


class BibliopolisReader(SRUReader):
    sru_url = 'http://jsru.kb.nl/sru/sru'
    sru_version = '1.2'
    HPB_LINK = 'http://hpb.cerl.org/record/{}'
    CATALOG_URIREF = URIRef(
        'https://edpop.hum.uu.nl/readers/bibliopolis'
    )
    IRI_PREFIX = "https://edpop.hum.uu.nl/readers/bibliopolis/"

    def __init__(self):
        super().__init__()
        self.additional_params = {
            'x-collection': 'Bibliopolis'
        }

    def _convert_record(self, sruthirecord: dict) -> Record:
        record = Record(from_reader=self.__class__)
        record.data = sruthirecord
        # TODO: extract fields
        return record

    @classmethod
    def transform_query(cls, query: str) -> str:
        return query
